#pragma once

#include "OJPH/T1OJPH.h"
#include "OJPH/PostT1DecompressFiltersOJPH.h"
#include "OJPH/QuantizerOJPH.h"

#include "OpenHTJ2K/T1OpenHTJ2K.h"
#include "OpenHTJ2K/PostT1DecompressFiltersOpenHTJ2K.h"
#include "OpenHTJ2K/QuantizerOpenHTJ2K.h"

//#define OPENHTJ2K
